@@ -23,10 +23,11 @@
 
 #include <ft2build.h>
 #include FT_FREETYPE_H
-#include <glib.h>
+#include <gio/gio.h>
 
 gchar * font_utils_get_font_name (FT_Face face);
 gchar * font_utils_get_font_name_for_file (FT_Library library,
-                                           const gchar *path);
+                                           GFile *file,
+                                           gint face_index);
 
 #endif /* __FONT_UTILS_H__ */
