@@ -42,7 +42,9 @@ enum {
   NUM_SIGNALS
 };
 
-struct _SushiFontWidgetPrivate {
+struct _SushiFontWidget {
+  GtkDrawingArea parent_instance;
+
   gchar *uri;
   gint face_index;
 
@@ -62,7 +64,7 @@ struct _SushiFontWidgetPrivate {
 static GParamSpec *properties[NUM_PROPERTIES] = { NULL, };
 static guint signals[NUM_SIGNALS] = { 0, };
 
-G_DEFINE_TYPE (SushiFontWidget, sushi_font_widget, GTK_TYPE_DRAWING_AREA);
+G_DEFINE_TYPE (SushiFontWidget, sushi_font_widget, GTK_TYPE_DRAWING_AREA)
 
 #define SURFACE_SIZE 4
 #define SECTION_SPACING 16
@@ -171,11 +173,11 @@ text_extents (cairo_t *cr,
               const char *text,
               cairo_text_extents_t *extents)
 {
-  cairo_glyph_t *glyphs;
+  g_autofree cairo_glyph_t *glyphs = NULL;
   gint num_glyphs;
+
   text_to_glyphs (cr, text, &glyphs, &num_glyphs);
   cairo_glyph_extents (cr, glyphs, num_glyphs, extents);
-  g_free (glyphs);
 }
 
 /* adapted from gnome-utils:font-viewer/font-view.c
@@ -192,9 +194,9 @@ draw_string (SushiFontWidget *self,
 	     const gchar *text,
 	     gint *pos_y)
 {
+  g_autofree cairo_glyph_t *glyphs = NULL;
   cairo_font_extents_t font_extents;
   cairo_text_extents_t extents;
-  cairo_glyph_t *glyphs;
   GtkTextDirection text_dir;
   gint pos_x;
   gint num_glyphs;
@@ -225,8 +227,6 @@ draw_string (SushiFontWidget *self,
   cairo_move_to (cr, pos_x, *pos_y);
   cairo_show_glyphs (cr, glyphs, num_glyphs);
 
-  g_free (glyphs);
-
   *pos_y += LINE_SPACING / 2;
 }
 
@@ -234,33 +234,25 @@ static gboolean
 check_font_contain_text (FT_Face face,
                          const gchar *text)
 {
-  gunichar *string;
+  g_autofree gunichar *string = NULL;
   glong len, idx;
-  gboolean retval = TRUE;
 
   string = g_utf8_to_ucs4_fast (text, -1, &len);
-
-  FT_Select_Charmap (face, FT_ENCODING_UNICODE);
-
   for (idx = 0; idx < len; idx++) {
     gunichar c = string[idx];
 
-    if (!FT_Get_Char_Index (face, c)) {
-      retval = FALSE;
-      break;
-    }
+    if (!FT_Get_Char_Index (face, c))
+      return FALSE;
   }
 
-  g_free (string);
-
-  return retval;
+  return TRUE;
 }
 
 static gchar *
 build_charlist_for_face (FT_Face face,
                          gint *length)
 {
-  GString *string;
+  g_autoptr(GString) string = NULL;
   gulong c;
   guint glyph;
   gint total_chars = 0;
@@ -278,16 +270,16 @@ build_charlist_for_face (FT_Face face,
   if (length)
     *length = total_chars;
 
-  return g_string_free (string, FALSE);
+  return g_strdup (string->str);
 }
 
 static gchar *
 random_string_from_available_chars (FT_Face face,
                                     gint n_chars)
 {
-  gchar *chars;
+  g_autofree gchar *chars = NULL;
+  g_autoptr(GString) retval = NULL;
   gint idx, rand, total_chars;
-  GString *retval;
   gchar *ptr, *end;
 
   idx = 0;
@@ -296,6 +288,9 @@ random_string_from_available_chars (FT_Face face,
   if (total_chars == 0)
     return NULL;
 
+  if (total_chars <= n_chars)
+    return g_steal_pointer (&chars);
+
   retval = g_string_new (NULL);
 
   while (idx < n_chars) {
@@ -308,7 +303,7 @@ random_string_from_available_chars (FT_Face face,
     idx++;
   }
 
-  return g_string_free (retval, FALSE);
+  return g_strdup (retval->str);
 }
 
 static gboolean
@@ -318,60 +313,73 @@ set_pango_sample_string (SushiFontWidget *self)
   gboolean retval = FALSE;
 
   sample_string = pango_language_get_sample_string (pango_language_from_string (NULL));
-  if (check_font_contain_text (self->priv->face, sample_string))
+  if (check_font_contain_text (self->face, sample_string))
     retval = TRUE;
 
   if (!retval) {
     sample_string = pango_language_get_sample_string (pango_language_from_string ("C"));
-    if (check_font_contain_text (self->priv->face, sample_string))
+    if (check_font_contain_text (self->face, sample_string))
       retval = TRUE;
   }
 
-  if (retval) {
-    g_free (self->priv->sample_string);
-    self->priv->sample_string = g_strdup (sample_string);
-  }
+  g_clear_pointer (&self->sample_string, g_free);
+
+  if (retval)
+    self->sample_string = g_strdup (sample_string);
 
   return retval;
 }
 
+static void
+select_best_charmap (SushiFontWidget *self)
+{
+  gchar *chars;
+  gint idx, n_chars;
+
+  if (FT_Select_Charmap (self->face, FT_ENCODING_UNICODE) == 0)
+    return;
+
+  for (idx = 0; idx < self->face->num_charmaps; idx++) {
+    if (FT_Set_Charmap (self->face, self->face->charmaps[idx]) != 0)
+      continue;
+
+    chars = build_charlist_for_face (self->face, &n_chars);
+    g_free (chars);
+
+    if (n_chars > 0)
+      break;
+  }
+}
+
 static void
 build_strings_for_face (SushiFontWidget *self)
 {
+  select_best_charmap (self);
+
   /* if we don't have lowercase/uppercase/punctuation text in the face,
    * we omit it directly, and render a random text below.
    */
-  if (check_font_contain_text (self->priv->face, lowercase_text_stock))
-    self->priv->lowercase_text = lowercase_text_stock;
+  if (check_font_contain_text (self->face, lowercase_text_stock))
+    self->lowercase_text = lowercase_text_stock;
   else
-    self->priv->lowercase_text = NULL;
+    self->lowercase_text = NULL;
 
-  if (check_font_contain_text (self->priv->face, uppercase_text_stock))
-    self->priv->uppercase_text = uppercase_text_stock;
+  if (check_font_contain_text (self->face, uppercase_text_stock))
+    self->uppercase_text = uppercase_text_stock;
   else
-    self->priv->uppercase_text = NULL;
+    self->uppercase_text = NULL;
 
-  if (check_font_contain_text (self->priv->face, punctuation_text_stock))
-    self->priv->punctuation_text = punctuation_text_stock;
+  if (check_font_contain_text (self->face, punctuation_text_stock))
+    self->punctuation_text = punctuation_text_stock;
   else
-    self->priv->punctuation_text = NULL;
+    self->punctuation_text = NULL;
 
   if (!set_pango_sample_string (self))
-    self->priv->sample_string = random_string_from_available_chars (self->priv->face, 36);
+    self->sample_string = random_string_from_available_chars (self->face, 36);
 
-  g_free (self->priv->font_name);
-  self->priv->font_name = NULL;
-
-  if (self->priv->face->family_name != NULL) {
-    gchar *font_name = 
-      g_strconcat (self->priv->face->family_name, " ",
-                   self->priv->face->style_name, NULL);
-
-    if (check_font_contain_text (self->priv->face, font_name))
-      self->priv->font_name = font_name;
-    else
-      g_free (font_name);
-  }
+  g_free (self->font_name);
+  self->font_name = g_strconcat (self->face->family_name, " ",
+                                 self->face->style_name, NULL);
 }
 
 static gint *
@@ -437,15 +445,15 @@ sushi_font_widget_size_request (GtkWidget *drawing_area,
                                 gint *min_height)
 {
   SushiFontWidget *self = SUSHI_FONT_WIDGET (drawing_area);
-  SushiFontWidgetPrivate *priv = self->priv;
   gint i, pixmap_width, pixmap_height;
   cairo_text_extents_t extents;
   cairo_font_extents_t font_extents;
   cairo_font_face_t *font;
-  gint *sizes = NULL, n_sizes, alpha_size, title_size;
+  g_autofree gint *sizes = NULL;
+  gint n_sizes, alpha_size, title_size;
   cairo_t *cr;
   cairo_surface_t *surface;
-  FT_Face face = priv->face;
+  FT_Face face = self->face;
   GtkStyleContext *context;
   GtkStateFlags state;
   GtkBorder padding;
@@ -478,50 +486,53 @@ sushi_font_widget_size_request (GtkWidget *drawing_area,
   pixmap_height = padding.top + padding.bottom;
 
   font = cairo_ft_font_face_create_for_ft_face (face, 0);
-  cairo_set_font_face (cr, font);
-  cairo_font_face_destroy (font);
 
-  if (self->priv->font_name != NULL) {
-      cairo_set_font_size (cr, title_size);
-      cairo_font_extents (cr, &font_extents);
-      text_extents (cr, self->priv->font_name, &extents);
-      pixmap_height += font_extents.ascent + font_extents.descent +
-        extents.y_advance + LINE_SPACING;
-      pixmap_width = MAX (pixmap_width, extents.width + padding.left + padding.right);
-  }
+  if (check_font_contain_text (face, self->font_name))
+    cairo_set_font_face (cr, font);
+  else
+    cairo_set_font_face (cr, NULL);
+
+  cairo_set_font_size (cr, title_size);
+  cairo_font_extents (cr, &font_extents);
+  text_extents (cr, self->font_name, &extents);
+  pixmap_height += font_extents.ascent + font_extents.descent +
+    extents.y_advance + LINE_SPACING;
+  pixmap_width = MAX (pixmap_width, extents.width + padding.left + padding.right);
 
   pixmap_height += SECTION_SPACING / 2;
+
+  cairo_set_font_face (cr, font);
   cairo_set_font_size (cr, alpha_size);
   cairo_font_extents (cr, &font_extents);
 
-  if (self->priv->lowercase_text != NULL) {
-    text_extents (cr, self->priv->lowercase_text, &extents);
+  if (self->lowercase_text != NULL) {
+    text_extents (cr, self->lowercase_text, &extents);
     pixmap_height += font_extents.ascent + font_extents.descent + 
       extents.y_advance + LINE_SPACING;
     pixmap_width = MAX (pixmap_width, extents.width + padding.left + padding.right);
   }
 
-  if (self->priv->uppercase_text != NULL) {
-    text_extents (cr, self->priv->uppercase_text, &extents);
+  if (self->uppercase_text != NULL) {
+    text_extents (cr, self->uppercase_text, &extents);
     pixmap_height += font_extents.ascent + font_extents.descent +
       extents.y_advance + LINE_SPACING;
     pixmap_width = MAX (pixmap_width, extents.width + padding.left + padding.right);
   }
 
-  if (self->priv->punctuation_text != NULL) {
-    text_extents (cr, self->priv->punctuation_text, &extents);
+  if (self->punctuation_text != NULL) {
+    text_extents (cr, self->punctuation_text, &extents);
     pixmap_height += font_extents.ascent + font_extents.descent +
       extents.y_advance + LINE_SPACING;
     pixmap_width = MAX (pixmap_width, extents.width + padding.left + padding.right);
   }
 
-  if (self->priv->sample_string != NULL) {
+  if (self->sample_string != NULL) {
     pixmap_height += SECTION_SPACING;
 
     for (i = 0; i < n_sizes; i++) {
       cairo_set_font_size (cr, sizes[i]);
       cairo_font_extents (cr, &font_extents);
-      text_extents (cr, self->priv->sample_string, &extents);
+      text_extents (cr, self->sample_string, &extents);
       pixmap_height += font_extents.ascent + font_extents.descent +
         extents.y_advance + LINE_SPACING;
       pixmap_width = MAX (pixmap_width, extents.width + padding.left + padding.right);
@@ -543,8 +554,8 @@ sushi_font_widget_size_request (GtkWidget *drawing_area,
     *height = pixmap_height;
 
   cairo_destroy (cr);
+  cairo_font_face_destroy (font);
   cairo_surface_destroy (surface);
-  g_free (sizes);
 }
 
 static void
@@ -556,7 +567,8 @@ sushi_font_widget_get_preferred_width (GtkWidget *drawing_area,
 
   sushi_font_widget_size_request (drawing_area, &width, NULL, NULL);
 
-  *minimum_width = *natural_width = width;
+  *minimum_width = 0;
+  *natural_width = width;
 }
 
 static void
@@ -577,10 +589,10 @@ sushi_font_widget_draw (GtkWidget *drawing_area,
                         cairo_t *cr)
 {
   SushiFontWidget *self = SUSHI_FONT_WIDGET (drawing_area);
-  SushiFontWidgetPrivate *priv = self->priv;
-  gint *sizes = NULL, n_sizes, alpha_size, title_size, pos_y = 0, i;
-  cairo_font_face_t *font;
-  FT_Face face = priv->face;
+  g_autofree gint *sizes = NULL;
+  gint n_sizes, alpha_size, title_size, pos_y = 0, i;
+  cairo_font_face_t *font = NULL;
+  FT_Face face = self->face;
   GtkStyleContext *context;
   GdkRGBA color;
   GtkBorder padding;
@@ -588,7 +600,7 @@ sushi_font_widget_draw (GtkWidget *drawing_area,
   gint allocated_width, allocated_height;
 
   if (face == NULL)
-    goto end;
+    return FALSE;
 
   context = gtk_widget_get_style_context (drawing_area);
   state = gtk_style_context_get_state (context);
@@ -607,34 +619,35 @@ sushi_font_widget_draw (GtkWidget *drawing_area,
   sizes = build_sizes_table (face, &n_sizes, &alpha_size, &title_size);
 
   font = cairo_ft_font_face_create_for_ft_face (face, 0);
-  cairo_set_font_face (cr, font);
-  cairo_font_face_destroy (font);
 
   /* draw text */
+  if (check_font_contain_text (face, self->font_name))
+    cairo_set_font_face (cr, font);
+  else
+    cairo_set_font_face (cr, NULL);
 
-  if (self->priv->font_name != NULL) {
-    cairo_set_font_size (cr, title_size);
-    draw_string (self, cr, padding, self->priv->font_name, &pos_y);
-  }
+  cairo_set_font_size (cr, title_size);
+  draw_string (self, cr, padding, self->font_name, &pos_y);
 
   if (pos_y > allocated_height)
     goto end;
 
   pos_y += SECTION_SPACING / 2;
+  cairo_set_font_face (cr, font);
   cairo_set_font_size (cr, alpha_size);
 
-  if (self->priv->lowercase_text != NULL)
-    draw_string (self, cr, padding, self->priv->lowercase_text, &pos_y);
+  if (self->lowercase_text != NULL)
+    draw_string (self, cr, padding, self->lowercase_text, &pos_y);
   if (pos_y > allocated_height)
     goto end;
 
-  if (self->priv->uppercase_text != NULL)
-    draw_string (self, cr, padding, self->priv->uppercase_text, &pos_y);
+  if (self->uppercase_text != NULL)
+    draw_string (self, cr, padding, self->uppercase_text, &pos_y);
   if (pos_y > allocated_height)
     goto end;
 
-  if (self->priv->punctuation_text != NULL)
-    draw_string (self, cr, padding, self->priv->punctuation_text, &pos_y);
+  if (self->punctuation_text != NULL)
+    draw_string (self, cr, padding, self->punctuation_text, &pos_y);
   if (pos_y > allocated_height)
     goto end;
 
@@ -642,14 +655,14 @@ sushi_font_widget_draw (GtkWidget *drawing_area,
 
   for (i = 0; i < n_sizes; i++) {
     cairo_set_font_size (cr, sizes[i]);
-    if (self->priv->sample_string !=  NULL)
-      draw_string (self, cr, padding, self->priv->sample_string, &pos_y);
+    if (self->sample_string !=  NULL)
+      draw_string (self, cr, padding, self->sample_string, &pos_y);
     if (pos_y > allocated_height)
       break;
   }
 
  end:
-  g_free (sizes);
+  cairo_font_face_destroy (font);
 
   return FALSE;
 }
@@ -660,17 +673,16 @@ font_face_async_ready_cb (GObject *object,
                           gpointer user_data)
 {
   SushiFontWidget *self = user_data;
-  GError *error = NULL;
+  g_autoptr(GError) error = NULL;
 
-  self->priv->face =
+  self->face =
     sushi_new_ft_face_from_uri_finish (result,
-                                       &self->priv->face_contents,
+                                       &self->face_contents,
                                        &error);
 
   if (error != NULL) {
-    g_signal_emit (self, signals[ERROR], 0, error->message);
+    g_signal_emit (self, signals[ERROR], 0, error);
     g_print ("Can't load the font face: %s\n", error->message);
-    g_error_free (error);
 
     return;
   }
@@ -684,9 +696,9 @@ font_face_async_ready_cb (GObject *object,
 void
 sushi_font_widget_load (SushiFontWidget *self)
 {
-  sushi_new_ft_face_from_uri_async (self->priv->library,
-                                    self->priv->uri,
-                                    self->priv->face_index,
+  sushi_new_ft_face_from_uri_async (self->library,
+                                    self->uri,
+                                    self->face_index,
                                     font_face_async_ready_cb,
                                     self);
 }
@@ -694,13 +706,7 @@ sushi_font_widget_load (SushiFontWidget *self)
 static void
 sushi_font_widget_init (SushiFontWidget *self)
 {
-  FT_Error err;
-
-  self->priv = G_TYPE_INSTANCE_GET_PRIVATE (self, SUSHI_TYPE_FONT_WIDGET,
-                                            SushiFontWidgetPrivate);
-
-  self->priv->face = NULL;
-  err = FT_Init_FreeType (&self->priv->library);
+  FT_Error err = FT_Init_FreeType (&self->library);
 
   if (err != FT_Err_Ok)
     g_error ("Unable to initialize FreeType");
@@ -719,10 +725,10 @@ sushi_font_widget_get_property (GObject *object,
 
   switch (prop_id) {
   case PROP_URI:
-    g_value_set_string (value, self->priv->uri);
+    g_value_set_string (value, self->uri);
     break;
   case PROP_FACE_INDEX:
-    g_value_set_int (value, self->priv->face_index);
+    g_value_set_int (value, self->face_index);
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -740,10 +746,10 @@ sushi_font_widget_set_property (GObject *object,
 
   switch (prop_id) {
   case PROP_URI:
-    self->priv->uri = g_value_dup_string (value);
+    self->uri = g_value_dup_string (value);
     break;
   case PROP_FACE_INDEX:
-    self->priv->face_index = g_value_get_int (value);
+    self->face_index = g_value_get_int (value);
     break;
   default:
     G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -756,20 +762,20 @@ sushi_font_widget_finalize (GObject *object)
 {
   SushiFontWidget *self = SUSHI_FONT_WIDGET (object);
 
-  g_free (self->priv->uri);
+  g_free (self->uri);
 
-  if (self->priv->face != NULL) {
-    FT_Done_Face (self->priv->face);
-    self->priv->face = NULL;
+  if (self->face != NULL) {
+    FT_Done_Face (self->face);
+    self->face = NULL;
   }
 
-  g_free (self->priv->font_name);
-  g_free (self->priv->sample_string);
-  g_free (self->priv->face_contents);
+  g_free (self->font_name);
+  g_free (self->sample_string);
+  g_free (self->face_contents);
 
-  if (self->priv->library != NULL) {
-    FT_Done_FreeType (self->priv->library);
-    self->priv->library = NULL;
+  if (self->library != NULL) {
+    FT_Done_FreeType (self->library);
+    self->library = NULL;
   }
 
   G_OBJECT_CLASS (sushi_font_widget_parent_class)->finalize (object);
@@ -823,10 +829,9 @@ sushi_font_widget_class_init (SushiFontWidgetClass *klass)
                   G_SIGNAL_RUN_FIRST,
                   0, NULL, NULL,
                   g_cclosure_marshal_VOID__STRING,
-                  G_TYPE_NONE, 1, G_TYPE_STRING);
+                  G_TYPE_NONE, 1, G_TYPE_ERROR);
 
   g_object_class_install_properties (oclass, NUM_PROPERTIES, properties);
-  g_type_class_add_private (klass, sizeof (SushiFontWidgetPrivate));
 }
 
 SushiFontWidget *
@@ -845,11 +850,11 @@ sushi_font_widget_new (const gchar *uri, gint face_index)
 FT_Face
 sushi_font_widget_get_ft_face (SushiFontWidget *self)
 {
-  return self->priv->face;
+  return self->face;
 }
 
 const gchar *
 sushi_font_widget_get_uri (SushiFontWidget *self)
 {
-  return self->priv->uri;
+  return self->uri;
 }
